% The code below was modified based on the code provided by Geoff Hinton 
% and Ruslan Salakhutdinov (Science paper: Reducing the Dimensionality
% Of Data with Neural Networks)

% This program trains Restricted Boltzmann Machine in which
% visible, binary, stochastic pixels are connected to
% hidden, binary, stochastic feature detectors using symmetrically
% weighted connections. Learning is done with 1-step Contrastive Divergence. 

% The program assumes that the following variables are set externally:
% maxepoch  -- maximum number of epochs
% numhid    -- number of hidden units 
% batchdata -- the data that is divided into batches (numcases numdims numbatches)
% restart   -- set to 1 if learning starts from beginning 

epsilonw      = 0.1;   % Learning rate for weights between visible units and hidden units 
epsilonvb     = 0.1;   % Learning rate for biases of visible units 
epsilonhb     = 0.1;   % Learning rate for biases of hidden units 
weightcost  = 0.0002;   
initialmomentum  = 0.5;
finalmomentum    = 0.9;

[numcases numdims numbatches]=size(batchdata);

if restart ==1,
   restart=0;
   epoch=1;

% Initializing symmetric weights and biases. 
   vishid     = 0.1*randn(numdims, numhid);
   hidbiases  = zeros(1,numhid);
   visbiases  = zeros(1,numdims);
   poshidprobs = zeros(numcases,numhid);
   neghidprobs = zeros(numcases,numhid);
   posprods    = zeros(numdims,numhid);
   negprods    = zeros(numdims,numhid);
   vishidinc  = zeros(numdims,numhid);
   hidbiasinc = zeros(1,numhid);
   visbiasinc = zeros(1,numdims);
   batchposhidprobs=zeros(numcases,numhid,numbatches);
end

for epoch = epoch:maxepoch,
    fprintf(1,'epoch %d\r',epoch); 
    errsum=0;
    for batch = 1:numbatches,
        fprintf(1,'epoch %d batch %d\r',epoch,batch); 

        %%%%%%%%% START POSITIVE PHASE %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        data = batchdata(:,:,batch);
        % P(Hj|Vi)=sigmoid(Vi*Wij+bj)
        poshidprobs = 1./(1 + exp(-data*vishid - repmat(hidbiases,numcases,1)));
        % dim(poshidprobs)=cases x hidden
        batchposhidprobs(:,:,batch)=poshidprobs;
        posprods    = data' * poshidprobs;
        % dim(posprods)=visible x hidden; dim(data')=visible x cases
        poshidact   = sum(poshidprobs);
        % dim(poshidact)=1 x hidden
        posvisact = sum(data);
        % dim(posvisact)=1 x visible
 
        %%%%% END OF POSITIVE PHASE  %%%%%
        poshidstates = poshidprobs > rand(numcases,numhid);

        %%%%% START NEGATIVE PHASE  %%%%%%
        negdata = 1./(1 + exp(-poshidstates*vishid' - repmat(visbiases,numcases,1)));
        % P(Vi|Hj)=sigmoid(Hj*Wij'+ai) 
        neghidprobs = 1./(1 + exp(-negdata*vishid - repmat(hidbiases,numcases,1)));    
        negprods  = negdata'*neghidprobs;
        neghidact = sum(neghidprobs);
        negvisact = sum(negdata); 

        %%%%% END OF NEGATIVE PHASE %%%%%
        err= sum(sum( (data-negdata).^2 ));
        errsum = err + errsum;

        if epoch>5,
           momentum=finalmomentum;
        else
           momentum=initialmomentum;
        end;

        %%%%% UPDATE WEIGHTS AND BIASES %%%%%
        vishidinc = momentum*vishidinc + ...
                    epsilonw*( (posprods-negprods)/numcases - weightcost*vishid);
        % posprods is the sum of visible*hidden through numcases. Therefore, we need to divide it by the number of cases => (posprods-negprods)/numcases
        visbiasinc = momentum*visbiasinc + (epsilonvb/numcases)*(posvisact-negvisact);
        % posvisact is the sum of visible through numcases. Therefore, we need to divide it by the number of cases => (posvisact-negvisact)/numcases
        hidbiasinc = momentum*hidbiasinc + (epsilonhb/numcases)*(poshidact-neghidact);
        % posvisact is the sum of hidden through numcases. Therefore, we need to divide it by the number of cases => (poshidact-neghidact)/numcases
    
        vishid = vishid + vishidinc;
        visbiases = visbiases + visbiasinc;
        hidbiases = hidbiases + hidbiasinc;

        %%%%% END OF UPDATES %%%%%

    end
    fprintf(1, 'epoch %4i error %6.1f  \n', epoch, errsum); 
end;
