% The code below was modified based on the code provided by Geoff Hinton 
% and Ruslan Salakhutdinov (Science paper: Reducing the Dimensionality
% Of Data with Neural Networks) 

% The difference between this rbm with interaction and rbm is the addition 
% of the edges between certain visible units. 

% The program assumes that the following variables are set externally:
% maxepoch  -- maximum number of epochs
% numhid    -- number of hidden units 
% batchdata -- the data that is divided into batches (numcases numdims numbatches)
% restart   -- set to 1 if learning starts from beginning 

epsilonw      = 0.1;   % Learning rate for weights between visible units and hidden units
epsilonvb     = 0.1;   % Learning rate for biases of visible units 
epsilonhb     = 0.1;   % Learning rate for biases of hidden units 
epsilonvv     = 1;     % Learning rate for weights between visible units themselves
weightcost  = 0.0002;   
initialmomentum  = 0.5;
finalmomentum    = 0.9;

[numcases numdims numbatches]=size(batchdata);

if restart ==1,
   restart=0;
   epoch=1;

   % Initializing symmetric weights and biases.
   vishid     = 0.1*randn(numdims, numhid);
   visvis     = 0.1*randn(numdims, numdims);
   %Don't allow the interaction between rat species and human species, 
   %the number of rat proteins is 16 and the number of human protein is also
   %16.
   visvis(1:16,17:32) =0;
   visvis(17:32,1:16) =0;
   %Ignore the influence of input protein on itself
   for number = 1:32,
       visvis(number,number) = 0;
   end
  
   hidbiases  = zeros(1,numhid);
   visbiases  = zeros(1,numdims);
   poshidprobs = zeros(numcases,numhid);
   neghidprobs = zeros(numcases,numhid);
   posprods    = zeros(numdims,numhid);
   negprods    = zeros(numdims,numhid);
   vishidinc  = zeros(numdims,numhid);
   visvisinc = zeros(numdims,numdims);
   hidbiasinc = zeros(1,numhid);
   visbiasinc = zeros(1,numdims);
   batchposhidprobs=zeros(numcases,numhid,numbatches);
   
end

for epoch = epoch:maxepoch,
    fprintf(1,'epoch %d\r',epoch); 
    errsum=0;
    for batch = 1:numbatches,
        fprintf(1,'epoch %d batch %d\r',epoch,batch); 

        %%%%% START POSITIVE PHASE %%%%%
        data = batchdata(:,:,batch);
        % P(Hj|Vi)=sigmoid(Vi*Wij+bj) 
        poshidprobs = 1./(1 + exp(-data*vishid - repmat(hidbiases,numcases,1)));
        % dim(poshidprobs)=cases x hidden 
        batchposhidprobs(:,:,batch) = poshidprobs;
        posprods = data' * poshidprobs;
        % dim(posprods)=visible x hidden; dim(data')=visible x cases
        poshidact = sum(poshidprobs);
        % dim(poshidact)=1 x hidden
        posvisact = sum(data);
        % dim(posvisact)=1 x visible
  
        %%%%% END OF POSITIVE PHASE  %%%%%
        poshidstates = poshidprobs > rand(numcases,numhid);

        %%%%% CALCULATE THE INTERACTIONS BETWEEN PROTEINS  %%%%%
        %add the for loop here to calculate the interaction between the visible nodes
        visinteraction = zeros(numcases,numdims); 
        for j = 1:1:numdims,
            visdata = batchdata;
            visvis(j,j) = 0;
            temp = visdata*visvis(:,j);
            visinteraction(:,j) =  temp;
        end

        %%%%% START NEGATIVE PHASE  %%%%%
        negdata = 1./(1 + exp(-poshidstates*vishid' - repmat(visbiases,numcases,1)-visinteraction));
        % P(Vi|Hj)=sigmoid(Hj*Wij'+ai) 
        neghidprobs = 1./(1 + exp(-negdata*vishid - repmat(hidbiases,numcases,1)));    
        negprods  = negdata'*neghidprobs;
        neghidact = sum(neghidprobs);
        negvisact = sum(negdata); 
  
        %%%%% END OF NEGATIVE PHASE %%%%%
        err= sum(sum( (data-negdata).^2 ));
        errsum = err + errsum;

        if epoch>5,
           momentum=finalmomentum;
        else
           momentum=initialmomentum;
        end;

        %%%%% UPDATE WEIGHTS AND BIASES %%%%%
        vishidinc = momentum*vishidinc + ...
                    epsilonw*( (posprods-negprods)/numcases - weightcost*vishid);
        % posprods is the sum of visible*hidden through numcases. Therefore, we need to divide it by the number of cases => (posprods-negprods)/numcases
        visbiasinc = momentum*visbiasinc + (epsilonvb/numcases)*(posvisact-negvisact);
        % posvisact is the sum of visible through numcases. Therefore, we need to divide it by the number of cases => (posvisact-negvisact)/numcases
        hidbiasinc = momentum*hidbiasinc + (epsilonhb/numcases)*(poshidact-neghidact);
        % posvisact is the sum of hidden through numcases. Therefore, we need to divide it by the number of cases => (poshidact-neghidact)/numcases
        visvisinc = momentum*visvisinc + (epsilonvv/numcases)*(transpose(batchdata)*batchdata - transpose(negdata)*negdata);
        % update of the visible-visible interaction

        vishid = vishid + vishidinc;
        visvis = visvis + visvisinc;
        visbiases = visbiases + visbiasinc;
        hidbiases = hidbiases + hidbiasinc;
        %Don't allow the interaction between rat species and human species and
        %the influence of protein on itself
        for number = 1:32,
            visvis(number,number) = 0;
        end
        visvis(1:16,17:32) =0;
        visvis(17:32,1:16) =0;
    
        %%%%% END OF UPDATES %%%%%
    end
    fprintf(1, 'epoch %4i error %6.1f  \n', epoch, errsum); 
end;